// Header File
// Created 05/02/05; 00:42:29

#ifndef __OPALE_H
#define	__OPALE_H

/*! Current version */
#define Opale_version "00.03.000"

/*! First primary constant to fix the highest number of tasks the kernel will handle.
		This is an internal constant, so you don't have to change it if you're not
		rebuilding the kernel itself. If you are, fix N so that X will be 2^N. 
		Because of the size of some tables, N can't be more than 3. */
#define _N_	3			
		
/*! Second constant to fix the highest number of tasks the kernel will handle. 
		This is an internal constant, so you don't have to change it if you're not
		rebuilding the kernel itself. It is the number of lines for all tasks tables.
		Y can take any value from 2 to 8. */
#define	_Y_ 8

/*! X is the number of rows for all tasks tables */
#define _X_ (1<<_N_)

#if _N_ > 3
#error "_N_ priority coord can't be more than 3"
#endif

#if _Y_ > 8
#error "_Y_ priority coord can't be more than 8"
#endif 

/*! Total number of tasks the kernel will be able to handle. Because of the size of some table, 
		it could be wise to take some standardize values for Y and N (such as 8/3 for 64 tasks, 4/2
		for 16 tasks and not try bigger than 8/3 (64 tasks).*/
#define NUM_TASKS (_X_*_Y_)

/*! Size of a table used to get the highest priority task in any task table. */
#define TO_LOWEST_SETTED_BIT_TABLE_SIZE ((_X_>_Y_)?(1<<_X_):(1<<_Y_))

/*! Size of the stack needed by the idle task (lowest priority task allocated by the kernel itself).
		This is the stack size needed to handle a single context switch. */
#define IDLE_TASK_STACK_SIZE 70

/*! Lowest priority. Used by the idle task. */
#define LOWEST_PRIORITY (NUM_TASKS-1)

/*! That macro enables to enter a critical section. While in a critical section, a task can't be interrupted
		by any other task or an interrupt since SR is set to 0x2700. It uses a custom trap 7. Never use this macro 
		if you are already in a critical section ! */
#define op_EnterCriticalSection() asm(" trap #7")

/*! Exits a critical section, and restores SR to the value it has before the call to op_EnterCriticalSection().
		Uses a custom trap 8. */
#define op_ExitCriticalSection() asm(" trap #8") 

/*! Structure describing a task. It is not useful to know exactly what's in since you don't have to
		initialize or read/modify the fields yourself if you are making a program using the kernel. */
typedef struct t_TASK {
	/*! Pointer to the stack allocated for the task. */
	char* stackPointer;
	/*! Number of remaining ticks before the task becomes ready to run.  */
	unsigned short tickDelay;
	/*! Pointer to the next sleeping task. Used by the kernel's ticker ISR. */
	struct t_TASK* nextSleeping;
	/*! Pointer to the previous sleeping task. Used by the kernel's ticker ISR */
	struct t_TASK* prevSleeping;
	/*! Identity of the task. It is although its priority. */
	unsigned char ident;
} t_TASK;

/*! Structure to handle waiting tasks (on an event, a tick, etc) */
typedef struct t_TASK_BLOCK {
	/*! Bit field. A bit is 1 when the byte of same index in tasksTable is not 0. */
	unsigned char rowIndex;
	/*! Table of bit field. Each bit is 1 when the tasks it handles is pending. */
	unsigned char tasksTable[ _Y_ ];
}	t_TASK_BLOCK;
	
/*! Structure that will contain some tables used by the kernel. Once again, it doesn't need
		to be initialized/modified by the user. */
typedef struct t_op_BUFFER {
	/*! Tasks ready to run */
	t_TASK_BLOCK readyTasks;
	/*! Table used for round-robin scheduling groups */
	t_TASK_BLOCK roundRobin;
	/*! Table of pointer to each of the tasks currently loaded in the kernel. */
	t_TASK* runningTasks[ NUM_TASKS ];	
	/*! Stack of the idle task (lowest priority task created by the kernel). */
	char idleTaskStack[ IDLE_TASK_STACK_SIZE ];
	/*! Table of values used to fasten computation. When given n, indexToLowestSettedBit[n] gives the
			index of the less significant bit setted. */
	unsigned char indexToLowestSettedBit[ TO_LOWEST_SETTED_BIT_TABLE_SIZE ];
} t_op_BUFFER;

extern unsigned long op_idleCounter;

void op_KernelInit( register t_op_BUFFER* buffer asm("%a0"), register unsigned char rotationMask asm("%d0") );
void op_KernelStart( void );
void op_KernelStop( void ) __attribute__ ((noreturn));

short op_TaskStart( register t_TASK* task asm("%a3"), register void(*taskEntryFunction)( register void* asm("%a0") ) asm("%a2"), register char* stackPointer asm("%a1"), register unsigned char ident asm("%d0"), register void* taskFunctionArgs asm("%a0") );
void op_TaskStop( void );
void op_TaskWaitForTicks( register unsigned short ticks asm("%d0") );

/*! Structure describing a semaphore event handler. Semaphores are objects that helps synchronyzation between tasks */
typedef struct t_SEMAPHORE {
	/*! Current value of the semaphore count */
	unsigned long	count;
	/*! Pending tasks */
	t_TASK_BLOCK waitingTasks;
} t_SEMAPHORE;

/*! Structure to describe a mailBox event handler. Mailboxes are objects that helps synchronyzation and data sharing between tasks */
typedef struct t_MAILBOX {
	/*! Pointer to the shared data. If null, no message has been sent yet. */
	void* message;
	/*! Pending tasks */
	t_TASK_BLOCK waitingTasks;
} t_MAILBOX;

/*! Structure to describe a message queue event handler. Messsage queues are objects that helps communication and data sharing between tasks */
typedef struct {
	/*! Pointer to the shared data buffer. */
	void** messageBuffer;
	/*! Pending tasks */
	t_TASK_BLOCK waitingTasks;
	/*! Index to the first message ready to read */
	unsigned short front;
	/*! Index of the first empty message slot on the back of the queue */
	unsigned short back;
	/*! Size off the queue */
	unsigned short bufferSize;
	/*! Number of messages in the queue */
	unsigned short bufferUsage;
} t_QUEUE;

void op_SemaphoreInit( register t_SEMAPHORE* semaphore asm("%a0"), register unsigned long initValue asm("%d0") );
void op_SemaphorePost( register t_SEMAPHORE* semaphore asm("%a0") );
void op_SemaphorePend( register t_SEMAPHORE* semaphore asm("%a0") );

void op_MailBoxInit( register t_MAILBOX* mailbox asm("%a0"), register void* initMessage asm("%d0") );
unsigned short op_MailBoxPost( register t_MAILBOX* mailBox asm("%a0"), register void* message asm("%d0") );
void* op_MailBoxPend( register t_MAILBOX* mailBox asm("%a0") );

void op_QueueInit( register t_QUEUE* queue asm("%a0"), register void* buffer asm("%d0"), register unsigned short size asm("%d1") );
unsigned short op_QueuePost( register t_QUEUE* queue asm("%a0"), register void* message asm("%d0") );
unsigned short op_QueuePostFront( register t_QUEUE* queue asm("%a0"), register void* message asm("%d0") );
void* op_QueuePend( register t_QUEUE* queue asm("%a0") );

#endif



/* end of file */
