// C Source File
// Created 05/02/05; 00:41:54

#include "opale_internal.h"

/*! indexToLowestSettedBit is a pointer to a table used to get the index of the lowest weighted setted bit
		of a number. Given the number, the value at indexToLowestSettedBit[ number ] is the index
		of the wanted bit. \n
		The size of the table is calculated in compilation time, since it doesn't have
		to be more than max( 2^X, 2^Y ) (cf TO_LOWEST_SETTED_BIT_TABLE_SIZE). \n
		The memory for the table is taken from the t_op_BUFFER given as an argument of op_KernelInit, and
		the table itself is initialized by taht function. \n \n
		
		Here is what would the table looks like fo 64 tasks: \n \n
\code		
indexToLowestSettedBit[ TO_LOWEST_SETTED_BIT_TABLE_SIZE ] =
{
	0, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	4, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	5, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	4, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	6, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	4, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	5, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	4, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	7, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	4, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	5, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	4, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	6, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	4, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	5, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0,
	4, 0, 1, 0,	2, 0, 1, 0,	3, 0, 1, 0,	2, 0, 1, 0
};
\endcode
*/
unsigned char* indexToLowestSettedBit;

unsigned short numberOfNestedISRs;
unsigned short numberOfNestedShedulerHalts;
unsigned short kernelStarted;

/*! Task Control Block for the idle task. */			
t_TASK idleTask;
/*! Pointer to the idle task's stack. Its depths is IDLE_TASK_STACK_SIZE bytes. */
char* idleTaskStack;

/*! Pointer to the array of running tasks. */
t_TASK** runningTasks;
/*! Pointer to the task block used for round-robin scheduling. */
t_TASK_BLOCK* roundRobin;
/*! pointer to the task block used for general purpose scheduling. */
t_TASK_BLOCK* readyTasks;

/** Inits the kernel. It is the first function you have to call before using any other
			function of Opale.
		@param buffer (a0.l) Pointer to a buffer of sizeof(t_op_BUFFER) bytes. Can be allocated in any writable memory
			(Heap or stack). When NUM_TASKS is big, be carefull to stack overflow. 
		@param rotationEnableMask (d0.b) MAsk enabling priority rotation for groups while scheduling. There are _Y_ groups of _X_ tasks each,
			when the correponding bit of rotationEnableMask is setted, all the tasks of the group have the same priority, and are run
			sequencially.
		*/
void op_KernelInit( register t_op_BUFFER* buffer asm("%a0"), register unsigned char rotationMask asm("%d0") )
{

	short i,j;
	
	indexToLowestSettedBit = buffer->indexToLowestSettedBit;
	readyTasks = &buffer->readyTasks;
	roundRobin = &buffer->roundRobin;
	idleTaskStack = buffer->idleTaskStack;
	runningTasks = buffer->runningTasks;

	for(i=1; i< TO_LOWEST_SETTED_BIT_TABLE_SIZE; indexToLowestSettedBit[i++]=j) for(j=0; !((i>>j)&1); j++); 

	/* inits the number of running ISR to 0 */
	numberOfNestedISRs = 0;

	/* inits the number of requests to stop the scheduler to 1 (halted) */
	numberOfNestedShedulerHalts = 1;
	kernelStarted = 0;

	/* no tasks are ready yet, so we clear the ready row index */
	readyTasks->rowIndex = 0;	
	roundRobin->rowIndex = rotationMask;
	
	/* inits each element of the readyTable to 0 (no active task) */
	
	for( i = 0; i<_Y_; )
	{
		readyTasks->tasksTable[ i ] = 0;
		roundRobin->tasksTable[ i++ ] = 0;
	}

	for( i = 0; i<NUM_TASKS ; )
	{
		runningTasks[ i++ ] = NULL;
	}

	InitIntVectors();

	op_TaskStart( &idleTask, IdleTaskFunction, &idleTaskStack[ IDLE_TASK_STACK_SIZE ], IDLE_TASK_IDENT, NULL );

}

/** Advance each sleeping task's counter and wake any task whose counter reaches zero.\n
		This function is called by the kernel tick interruption subroutine.
*/
void PerformDelay( void )
{

t_TASK* task = &idleTask;

	while( ( task = task->nextSleeping ) != NULL )
	{
		if( --(task->tickDelay) <= 0 )
		{

			addTaskToTable( readyTasks, task->ident );

			if( ( (task->prevSleeping)->nextSleeping = task->nextSleeping ) ) 
				(task->nextSleeping)->prevSleeping = task->prevSleeping;
		
		}
	}
}


static unsigned char rotate( unsigned char t, register unsigned char n asm("%d2") )
{
	return (((t>>n)|(t<<(_X_-n))));
}

/** The scheduler is used to find the task that is to be run
*/
void Scheduler( void ) 
{
	
	/*! 	Find the task with the highest priority ready to run, and make switch the context to run it
			if it isn't already running. \n \n

			Let's see the tools we'll use to do the job ... \n \n


	** indexToLowestSettedBitIndex : \n
\code
			static const unsigned char indexToLowestUnsettedBitIndex [ 16 ] = {	0, 0, 1, 0,
																																					2, 0, 1, 0,
																																					3, 0, 1, 0,
\endcode																																					2, 0, 1, 0 };
			This table converts the number you give as an index to the correspondinb less significant setted bit it contains. \n
			It works for a 4-bits index, since we have 16 tasks max. \n
			The first 8 values (index 0-7) are never used since there's always alt least the idle task runnig. \n

			ex : \n
					0b00001000 -> 3 \n
					0b00001101 -> 1 \n
					0b00001111 -> 0 \n
					etc... \n \n

	** readyTasksTable : \n \n
\code
			<-------- X -----------+
			  7 6 5 4   3 2 1 0    |
			--------------------+  |			It is a table of 4 bytes, but we use ontly the lowest 4 bits of those.
			| . . . . | 3 2 1 0 |  |	0		I draw on here the representation of the array with the bits correspounding
			| . . . . | 7 6 5 4 |  Y	1		to the tasks' priorities. When a bit is set, the task is ready and pending.
			| . . . . | B A 9 8 |  |	2		The 0 is the highest priority.
			| . . .	. | F E D C |  V	3		The table is never empty since the 15th task (idle) is always runnig !
\endcode
 \n
	** readyRowIndex : \n

			| . . . . | 3 2 1 0 | \n \n

			The bit nx of readyTasksRowIndex is set when the row nx of readyTasksTable is not 0. \n
			That means at least one task is pending in that row. \n
			That index cannot be 0 since there's always at least one task running (the idle task). \n \n

	** priority : \n \n

							   <-Y-> <-X-> \n
			| . . . . | 3 2 | 1 0 | \n \n

			A task priority is unique, and goes from 0 (highest) to 15 (lowest) \n
			Writen in binary, it can be used like that : \n \n

				bits 0-1 are the X index in the ready tasks table. \n
				bits 2-3 are the Y index. \n \n

			It is just because of how you write it in binary, and because the ready tasks table is 4*4. \n
			Priorities are also used as idents for the tasks. \n \n

			Now it becomes easy to get the highest priority task : \n \n

			indexToLowestSettedBitIndex[ readyTasksRowIndex ] gives y (the lowest index of a row containing pending tasks) \n
			indexToLowestSettedBitIndex[ readyTasksTable[ y ] ] gives X (index of the lowest setted bit of the lowest row containing setted bit, taht is: pending tasks) \n \n

			Here we have the priority / ident of the task: \n \n

			ident = ( y << 2 ) + x \n

			 */
		 
		/* x and y shall not be allocated on the stack ! */
		register unsigned char y asm("%d1") = indexToLowestSettedBit[ readyTasks->rowIndex ]; 
		register unsigned char x asm("%d0") ; 

		if( roundRobin->rowIndex & indexToSettedBit( y ) )
		{
			
			register unsigned char decalage asm("%d2") = roundRobin->tasksTable[ y ] + 1;

			x = ( indexToLowestSettedBit[ rotate( readyTasks->tasksTable[ y ], decalage ) ] + decalage ) & MASK_X;
			roundRobin->tasksTable[ y ] = x;
		}
		else
		{
			x = indexToLowestSettedBit[ readyTasks->tasksTable[ y ] ];
		}
		
		x += ( y << MASK_Y );
			
		/* perform context switch if this is not the current task */
		if( currentTask->ident != x )
		{
		
			register t_TASK* highestPriorityTaskReady asm("%a3") = runningTasks[ x ];
			/* context switch */
			ContextSwitch( highestPriorityTaskReady, &currentTask );
		}

	/* work done ... */

}

//
//void op_EnterISR( void )
//{
//	/*	increments the number of ISR in execution ... should be performed in an atomic assembly instruction
//			(I need to check the assembly output) */
//
//	numberOfNestedISRs ++;
//
//}


//void op_ExitISR( void )
//{
	/*	decrements the number of ISR in execution ... should also be performed in an atomic assembly instruction
			(I need to check the assembly output, too)
			If the number is back to , the we are back to a non-ISR function, we then call the scheduler
			(if it is enabled) to check if one of the ISR that just finished hasn't woken
			a high priority thread */

//	if( !( -- numberOfNestedISRs ) && ( numberOfNestedShedulerHalts ) )
//	{
//		op_Scheduler();
//	}
//}

