// C Source File
// Created 21/02/05; 19:16:17

#include "opale_internal.h"

/** General initialization for events (semaphores, mailboxes, queues).
		Set values of common fields of the events control blocks.
		op_SemaphoreInit and op_MailBoxInit are just redefinitions of this function. 
			op_QueueInit uses it too, but with complementary code.
		@param event (a0.l) Pointer to the t_EVENT to be initialized.
		@param initValue (d0.l) Value for the t_EVENT.value field. Could be the semaphore's counter, mailbox message or queue memory buffer.
		*/
static void EventInit( register t_EVENT* event asm("%a0"), register unsigned long initValue asm("%d0") )
{
	short i;																							/* just a counter */
	unsigned char* ptr = event->waitingTasks.tasksTable;	/* gets a pointer to the taskTable */
	
	event->waitingTasks.rowIndex = 0;											/* sets the t_EVENT.waitingTasks.rowIndex field to 0 (no task is waiting for the event */
	event->value = initValue;															/* sets the t_EVENT.value to the wanted value */
		
	for( i = 0; i<_Y_; i++ )															/* clear the waitingTasks table */
	{
		*ptr++ = 0;
	}
}

/**	Inits a Semaphore with an initial count. It is done so you just don't have to fill manually the fields of the t_SEMAPHORE object.
		@param semaphore (a0.l) Pointer to the t_SEMAPHORE object to initialyze.
		@param initValue (d0.l) Value to initialyze the semaphore with. Remember semaphore are counter and blocking when 0.
		*/
void op_SemaphoreInit( register t_SEMAPHORE* semaphore asm("%a0"), register unsigned long initValue asm("%d0") )
{
	EventInit( (t_EVENT*)semaphore, initValue );					/* simple call to EventInit with a cast */
}

/**	Inits a MailBox with an initial message. It is done so you just don't have to fill manually the fields of the t_MAILBOX object.
		@param mailbox (a0.l) Pointer to the t_MAILBOX object to initialyze.
		@param initMessage (d0.l) Pointer to initialyze the mailbox with.
		*/
void op_MailBoxInit( register t_MAILBOX* mailbox asm("%a0"), register void* initMessage asm("%d0") )
{
	EventInit( (t_EVENT*)mailbox, (unsigned long)initMessage );	/* simple call to EventInit with a cast */
}

/**	Inits a message queue with an initial message. It is done so you just don't have to fill manually the fields of the t_MAILBOX object.
		@param mailbox (a0.l) Pointer to the t_MAILBOX object to initialyze.
		@param initMessage (d0.l) Pointer to initialyze the mailbox with.
		*/
void op_QueueInit( register t_QUEUE* queue asm("%a0"), register void* buffer asm("%d0"), register unsigned short size asm("%d1") )
{
	EventInit( (t_EVENT*)queue, (unsigned long)buffer );	/* simple call to EventInit with a cast */

	queue->bufferSize = size;															/* set the t_QUEUE.bufferSize field to the wanted size (maximum number of messages the queue can handle) */
	queue->front = 0;																			/* set the t_QUEUE.front index to 0 */
	queue->back = 0;																			/* set the t_QUEUE.back index to 0 */
	queue->bufferUsage = 0;																/* set the t_QUEUE.bufferUsage field to 0 (current number of messages in the queue */
}

/**	Adds a task to a t_TASK_BLOCK. This could be either a table of waiting tasks, or sleeping tasks, or so.
		@param table (a0.l) Pointer to the table
		@param ident (d0.l) Ident of the task to insert
*/
void addTaskToTable( register t_TASK_BLOCK* table asm("%a0"), register unsigned char ident asm("%d0") ) 
{ 
	unsigned char x = identToX( ident ); 								/* get the X part of the ident of the task */
	unsigned char y = identToY( ident ); 								/* get the Y part of the ident of the task */
	
	poke_bset( &table->tasksTable[ y ], x );						/* set the task's bit in the correct line of the table */
	poke_bset( &table->rowIndex, y );										/* set the line's bit in the table index */
} 

/**	Remove a task from a t_TASK_BLOCK. This could be either a table of waiting tasks, or sleeping tasks, or so.
		@param table (a0.l) Pointer to the table
		@param ident (d0.l) Ident of the task to remove
*/
void removeTaskFromTable( register t_TASK_BLOCK* table asm("%a0"), register unsigned char ident asm("%d0")  ) 
{
	unsigned char x = identToX( ident ); 								/* get the X part of the ident of the task */
	unsigned char y = identToY( ident ); 								/* get the Y part of the ident of the task */
	
	poke_bclr( &table->tasksTable[ y ], x ); 						/* clear the task's bit in the correct line of the table */
	if( !table->tasksTable[ y ] ) 											/* check if there is anymore bit setted on this line */
		poke_bclr( &table->rowIndex, y ); 								/* if there is none, clears the line's bit in the table's index */
}

/** Tranfer a task from a waiting table to another. Used for making a task to sleep or to wake it.
		@param tableFrom (a0.l) Table of the event on which the table is waiting
		@param tableTO (a1.l) Table to transfer the task to
		@param ident (d0.l) Ident of the task to move from a table to another
*/
static void moveFromTable( register t_TASK_BLOCK* tableFrom asm("%a0"), register t_TASK_BLOCK* tableTo asm("%a1"), register unsigned char ident asm("d0") )
{
	addTaskToTable( tableTo, ident );					/* add the task to tableTo */
	removeTaskFromTable( tableFrom, ident );	/* remove it from tableFrom */

	op_Scheduler();														/* since ready tasks waiting could have changed, call scheduler */
}

/** Makes the highest priority task from a waiting table ready to run (transfers the task to the readyTasks table)
		@param event (a0.l) Table from which to get the task
*/
static void MakeTaskReady( register t_EVENT* event asm("%a0") )
{
	if( event->waitingTasks.rowIndex )				/* check if at least one task is waiting */
	{

	  unsigned char y = indexToLowestSettedBit[ event->waitingTasks.rowIndex ];
		unsigned char x = indexToLowestSettedBit[ event->waitingTasks.tasksTable[ y ] ];
		
		register unsigned char highestPriorityWaiting asm("%d0") = ( y << MASK_Y ) + x;		/* find the highest priority task waiting for the event */
		
		moveFromTable( &event->waitingTasks, readyTasks, highestPriorityWaiting );				/* removes the task from the waiting queue and adds the task to the ready list */
	}
}

/** Transfers the current task to the waiting table of the wanted event 
		@param	event (a0.l) Pointer to the t_EVENT in which table the task is to be put to wait
*/
static void WaitForEvent( register t_EVENT* event asm("%a0") )
{
	/* removes the task from the ready queue and adds the task to the waiting list */
	moveFromTable( readyTasks, &event->waitingTasks, currentTask->ident );
}

/**	Post to the semaphore. It will increment the semaphore's count, and made ready to run the highest priority task pending on it.
			The scheduler is then called.
		@param semaphore (a0.l) Semaphore to post to.
		*/
void op_SemaphorePost( register t_SEMAPHORE* semaphore asm("%a0") )
{

	op_EnterCriticalSection();
	
	semaphore->count++;	

	MakeTaskReady( (t_EVENT*)semaphore );

	op_ExitCriticalSection();
}

/** Pend on a semaphore. If the count is not 0, the function will just return. If the semaphore has already
			been taken (its count isn't 0), the task will be removed from the ready list and wait until
			some other task post to the semaphore, and no other task of higher priority is pending on it.
		@param semaphore (a0.l) Semaphore to pend on.
		*/
void op_SemaphorePend( register t_SEMAPHORE* semaphore asm("%a0") )
{
	op_EnterCriticalSection();
	
	if( semaphore->count == 0 )
	{
		WaitForEvent( (t_EVENT*)semaphore );
	}
	
	semaphore->count--;
	op_ExitCriticalSection();
}

/**	Post to the mailbox. It will fill the message field if it isn't already not NULL. If so, the function will return an error. 
			If tasks were waiting on the mailbox, the highest priority one is made ready to run and the scheduler is called.
		@param mailBox (a0.l) MailBox to post to.
		@param message (d0.l) Message to post.
		@return FALSE if a message is already in the mailbox (can't post), or TRUE on success.
		*/
unsigned short op_MailBoxPost( register t_MAILBOX* mailBox asm("%a0"), register void* message asm("%d0") )
{

	op_EnterCriticalSection();
	
	if( mailBox->message != NULL ) return FALSE;
	
	mailBox->message = message;
	
	MakeTaskReady( (t_EVENT*)mailBox );

	op_ExitCriticalSection();
	return TRUE;
}

/** Wait for a message to be posted in the mailbox. If a message already is in the mailbox when the function is called, 
			it will just return it. If there's no message, the task will be removed from the ready list
			and make ready again only when a message has been posted and the task is the highest priority one waiting
			for it.
		@param mailBox (a0.l) MailBox to pend on.
		@return Message posted on the mailbox.
		*/
void* op_MailBoxPend( register t_MAILBOX* mailBox asm("%a0") )
{
	void* message;

	op_EnterCriticalSection();
	
	if( mailBox->message == NULL )
	{
		WaitForEvent( (t_EVENT*)mailBox );	
	}

	message = mailBox->message;
	mailBox->message = NULL;
	
	op_ExitCriticalSection();
	
	return message;
}

/**	Post a message on the back of the message queue. It will add the message to the back of the queue if the queue isn't already full.
			If so, the function will return an error. 
			If tasks were waiting on the message queue, the highest priority one is made ready to run and the scheduler is called.
		@param queue (a0.l) Queue to post to.
		@param message (d0.l) Message to post.
		@return FALSE if the queue is full (can't post), or TRUE on success.
		*/
unsigned short op_QueuePost( register t_QUEUE* queue asm("%a0"), register void* message asm("%d0") )
{
	op_EnterCriticalSection();
	
	if( queue->bufferUsage >= queue->bufferSize ) return FALSE;
	
	queue->messageBuffer[ queue->back ] = message;
	queue->back = (queue->back >= queue->bufferSize)? 0: queue->back+1;
	queue->bufferUsage++;
	
	MakeTaskReady( (t_EVENT*)queue );

	op_ExitCriticalSection();
	return TRUE;
}

/**	Post a message on the front of the message queue. It will add the message to the front of the queue if the queue isn't already full.
			If so, the function will return an error. This is different from QueuePost since the messages are taken from the front of
			the queue on a QueuePend. So a message posted with QueuePostFronT will be read before one posted with QueuePost.
			If tasks were waiting on the message queue, the highest priority one is made ready to run and the scheduler is called.
		@param queue (a0.l) Queue to post to.
		@param message (d0.l) Message to post.
		@return FALSE if the queue is full (can't post), or TRUE on success.
		*/
unsigned short op_QueuePostFront( register t_QUEUE* queue asm("%a0"), register void* message asm("%d0") )
{
	op_EnterCriticalSection();
	
	if( queue->bufferUsage >= queue->bufferSize ) return FALSE;
	
	queue->front = ((queue->front == 0)? queue->bufferSize: queue->front) -1;

	queue->messageBuffer[ queue->front ] = message;
	queue->bufferUsage++;
	
	MakeTaskReady( (t_EVENT*)queue );

	op_ExitCriticalSection();
	return TRUE;
}

/** Wait for a message to be posted in the message queue. If one or more message already is in the queue when the function is called, 
			it will just return the one on the front of the queue. If there's no message, the task will be removed from the ready list
			and make ready again only when a message has been posted and the task is the highest priority one waiting
			for it.
		@param queue (a0.l) Queue to pend on.
		@return Message posted on the message queue.
		*/
void* op_QueuePend( register t_QUEUE* queue asm("%a0") )
{
	void* message;
	op_EnterCriticalSection();
	
	if( queue->bufferUsage == 0 )
	{
		WaitForEvent( (t_EVENT*)queue );	
	}

	message = queue->messageBuffer[ queue->front ];
	
	if( ++(queue->front) >= queue->bufferSize )
		queue->front = 0;

	queue->bufferUsage--;
	
	op_ExitCriticalSection();
	
	return message;
}

/* end of file */

