// C Source File
// Created 05/02/05; 00:42:03

#include "opale_internal.h"

/** Loads a task in the kernel. After the call to op_TaskStart, if the scheduler is activated, the highest
			priority task will be made ready to run. 
		@param task (a3.l) Pointer to a unique t_TASK structure.
		@param taskEntryFunction (a2.l) Pointer to the function of the task. That function should never return, and gets a pointer
			to its parameters via the a0 register. Prototype of the functin : void taskEntryFunction( register void* args asm("%a0")
		@param stackPointer (a1.l) Pointer to a buffer allocated to be the stack of the task's function. Be careful to give a buffer of a suffivier size.
		@param ident (d0.b) Identity of the function. It is also its priority. Remember 0 is the highest priority, and LOWEST_PRIORITY is reserved to the idle task.
			If a task of the same ident/priority has already been created, the current call to op_TaskStart will fail.
		@param taskFunctionArgs (a0.l) Pointer to a structure that will be given as a parameter to the function of the task.
		@return FALSE if the requested identity already exists, TRUE in any other case.
		*/
short op_TaskStart( register t_TASK* task asm("%a3"), register void(*taskEntryFunction)( register void* asm("%a0") ) asm("%a2"), register char* stackPointer asm("%a1"), register unsigned char ident asm("%d0"), register void* taskFunctionArgs asm("%a0") )
{

	if( runningTasks[ ident ] != NULL ) return FALSE;

	/* inits the fields of the task's structure */
	task->ident = ident;
	task->prevSleeping = NULL;
	task->nextSleeping = NULL;
	
	/* inits the task's stack */
	task->stackPointer = InitStack( taskFunctionArgs, stackPointer, taskEntryFunction );

	/* add the task to the list of tasks */
	op_EnterCriticalSection();

	runningTasks[ ident ] = task;
	addTaskToTable( readyTasks, ident );
		
	op_ExitCriticalSection();

	op_Scheduler();

	return TRUE;
}

/** Stops the current task. Use it at the end of a task function when you want to end it. Remember
			task functions shall not return, so if you want to quit one of those, use op_TaskStop 
		*/
void op_TaskStop( void )
{
unsigned char ident = currentTask->ident;
	
	op_EnterCriticalSection();
	
	runningTasks[ ident ] = NULL;
	removeTaskFromTable( readyTasks, ident );

	op_ExitCriticalSection();
	op_Scheduler();
	
}

/** Make a task wait for a number of ticks. A tick occurs each time the int 5 is called. The task will be made
			ready to run when the given time has expired and it is the highest priority task ready to run.
		@param ticks (d0.w) Number of ticks to wait for.
		*/
void op_TaskWaitForTicks( register unsigned short ticks asm("%d0") )
{

	currentTask->tickDelay = ticks;
	
	op_EnterCriticalSection();
	
	currentTask->prevSleeping = &idleTask;

	if( (currentTask->nextSleeping = idleTask.nextSleeping) )
		(idleTask.nextSleeping)->prevSleeping = currentTask;
			
	idleTask.nextSleeping = currentTask;

	removeTaskFromTable( readyTasks, currentTask->ident );
			
	op_ExitCriticalSection();			

	op_Scheduler();

}

/* end of file */
